from __future__ import annotations
import datetime
from decimal import Decimal
import json
from typing import Optional, List
from pydantic import BaseModel, ConfigDict, Field


from paymentsgate.enums import (
    Currencies,
    InvoiceTypes,
    Languages,
    Statuses,
    TTLUnits,
    CurrencyTypes,
    FeesStrategy,
    InvoiceDirection,
    CredentialsTypes,
)


class BaseRequestModel(BaseModel):
    model_config = ConfigDict(extra="forbid")


class BaseResponseModel(BaseModel):
    model_config = ConfigDict(extra="ignore")


class Credentials(BaseModel):
    account_id: str
    public_key: str
    private_key: Optional[str] = Field(default=None)
    merchant_id: Optional[str] = Field(default=None)
    project_id: Optional[str] = Field(default=None)

    @classmethod
    def fromFile(cls, filename):
        data = json.load(open(filename))
        return cls(**data)

    model_config = ConfigDict(extra="ignore")


class PayInFingerprintBrowserModel(BaseRequestModel):
    acceptHeader: str
    colorDepth: int
    language: str
    screenHeight: int
    screenWidth: int
    timezone: str
    userAgent: str
    javaEnabled: bool
    windowHeight: int
    windowWidth: int


class PayInFingerprintModel(BaseRequestModel):
    fingerprint: str
    ip: str
    country: str
    city: str
    state: str
    zip: str
    browser: Optional[PayInFingerprintBrowserModel]


class PayInModel(BaseRequestModel):
    amount: float  # decimals: 2
    currency: Currencies
    country: Optional[str]  # Country iso code
    invoiceId: Optional[str]  # idempotent key
    clientId: Optional[str]  # uniq client ref
    type: InvoiceTypes  # Invoice subtype, see documentation
    bankId: Optional[str]  # ID from bank list or NSPK id
    trusted: Optional[bool]
    successUrl: Optional[str]
    failUrl: Optional[str]
    backUrl: Optional[str]
    clientCard: Optional[str]
    clientName: Optional[str]
    fingerprint: Optional[PayInFingerprintModel]
    lang: Optional[Languages]
    sync: Optional[bool]  # sync h2h scheme, see documentation
    multiWidgetOptions: Optional[PayInMultiWidgetOptions]
    theme: Optional[str]  # personalized widget theme


class PayInResponseModel(BaseResponseModel):
    id: str
    status: Statuses
    type: InvoiceTypes
    url: Optional[str]
    deeplink: Optional[str]
    m10: Optional[str]
    cardholder: Optional[str]
    account: Optional[str]
    bankId: Optional[str]
    accountSubType: Optional[str]


class PayOutRecipientModel(BaseRequestModel):
    account_number: str | None = None  #  IBAN, Phone, Card, local bank account number, wallet number, etc'
    account_owner: str | None = None  # FirstName LastName or FirstName MiddleName LastName
    account_iban: str | None = None  # use only cases where iban is't primary  account id
    account_swift: str | None = None  # for swift transfers only
    account_phone: str | None = None  # additional recipient phone number, use only cases where phone is't primary  account id
    account_bic: str | None = None  # recipient bank id
    account_ewallet_name: str | None = None  # additional recipient wallet provider info
    account_email: str | None = None  # additional recipient email, use only cases where email is't primary account id
    account_bank_id: str | None = None  # recipient bankId (from API banks or RU NSPK id)
    account_internal_client_number: str | None = None  # Bank internal identifier used for method banktransferphp (Philippines)
    type: CredentialsTypes | None = None  # primary credential type


class PayOutModel(BaseRequestModel):
    currency: Currencies | None = None  # currency from, by default = usdt
    currencyTo: Currencies | None = None  # currency to, fiat only, if use quoteId - not required
    amount: Decimal | None = None  # decimals: 2, if use quoteId - not required
    invoiceId: str | None = None  # idempotent key
    clientId: str | None = None  # uniq client ref
    ttl: int | None = None
    ttl_unit: TTLUnits | None = None
    finalAmount: Decimal | None = None  # Optional, for pre-charge rate lock
    sender_name: str | None = None  # sender personal short data
    sender_personal: PayOutSenderModel | None = None
    baseCurrency: CurrencyTypes | None = None
    feesStrategy: FeesStrategy | None = None
    recipient: PayOutRecipientModel
    quoteId: str | None = None
    src_amount: str | None = None  # Optional, source amount in local currency for 2phase payout
    type: InvoiceTypes | None = None  # payout transaction scheme hint


class PayOutResponseModel(BaseResponseModel):
    id: str
    status: str


class GetQuoteModel(BaseRequestModel):
    currency_from: Currencies
    currency_to: Currencies
    amount: Decimal
    subtype: InvoiceTypes | None = None
    currency_original: Currencies | None = None


class QuoteEntity(BaseResponseModel):
    currencyFrom: Currencies
    currencyTo: Currencies
    pair: str
    rate: float


class GetQuoteResponseModel(BaseResponseModel):
    id: str
    finalAmount: Decimal
    direction: InvoiceDirection
    fullRate: Decimal
    fullRateReverse: Decimal
    fees: Decimal
    fees_percent: Decimal
    quotes: List[QuoteEntity]
    expiredAt: Optional[datetime.datetime] | None = None

    # deprecated
    currency_from: Optional[CurrencyModel] = Field(default=None)
    currency_to: Optional[CurrencyModel] = Field(default=None)
    currency_middle: Optional[CurrencyModel] = Field(default=None)
    rate1: Optional[float] = Field(default=None)
    rate2: Optional[float] = Field(default=None)
    rate3: Optional[float] = Field(default=None)
    net_amount: Optional[float] = Field(default=None)
    metadata: Optional[object] = Field(default=None)


class DepositAddressResponseModel(BaseResponseModel):
    currency: Currencies
    address: str
    expiredAt: datetime.datetime


class CurrencyModel(BaseResponseModel):
    _id: str
    type: CurrencyTypes
    code: Currencies
    symbol: str
    label: Optional[str] = Field(default=None)
    decimal: int
    countryCode: Optional[str] = Field(default=None)
    countryName: Optional[str] = Field(default=None)


class BankModel(BaseResponseModel):
    name: str
    title: str
    currency: Currencies
    fpsId: str


class InvoiceStatusModel(BaseResponseModel):
    name: Statuses
    createdAt: datetime.datetime
    updatedAt: datetime.datetime


class InvoiceAmountModel(BaseResponseModel):
    crypto: float
    fiat: float
    fiat_net: float


class InvoiceMetadataModel(BaseResponseModel):
    invoiceId: Optional[str]
    clientId: Optional[str]
    fiatAmount: Optional[float]


class InvoiceModel(BaseResponseModel):
    id: str | None = Field(..., alias='_id')
    orderId: str
    projectId: str
    currencyFrom: CurrencyModel
    currencyTo: CurrencyModel
    direction: InvoiceDirection
    amount: float
    status: InvoiceStatusModel
    amounts: InvoiceAmountModel
    metadata: InvoiceMetadataModel
    receiptUrls: List[str]
    isExpired: bool
    createdAt: datetime.datetime | None = None
    updatedAt: datetime.datetime | None = None
    expiredAt: datetime.datetime | None = None


class AssetsAccountModel(BaseResponseModel):
    currency: CurrencyModel
    total: float
    pending: float
    available: float


class AssetsResponseModel(BaseResponseModel):
    assets: List[AssetsAccountModel]


class PayInMultiWidgetOptions(BaseRequestModel):
    offerAmount: Optional[bool]  # show amount select from best offers
    elqrBanks: Optional[str]  # elqr bank list


class PayOutSenderModel(BaseRequestModel):
    name: Optional[str]
    birthday: Optional[str]
    phone: Optional[str]
    passport: Optional[str]


class PayOutTlvRequestModel(BaseRequestModel):
    quoteId: str  # ID from /fx/tlv response
    invoiceId: Optional[str]
    clientId: Optional[str]
    sender_personal: Optional[PayOutSenderModel]


class GetQuoteTlv(BaseRequestModel):
    data: str

class TLVExtended(BaseResponseModel):
    merchant: str | None = None # MerchantName
    logo: str | None = None # MerchantLogo or merchant MCC logo
    city: str | None = None # Merchant city
    merchantId: str | None = None # uniq merchant id 
    zip: str | None = None # merchant address zip code
    qrRefId: str | None = None # uniq QR code reference id 
    invoiceId: str | None = None # Merchant invoiceId 
    merchantBank: str | None = None # Merchant bank name
    merchantIban: str | None = None # merchant iban
    merchantBankLogo: str | None = None # merchant bank logo

class QuoteTlvResponse(BaseResponseModel):
    id: str
    amount: float  # fiat local amount
    amountCrypto: float  # total crypto amount inc. fees
    currencyCode: Currencies  # local currency
    feeInCrypto: float  # total fee in crypto
    feePercent: float  # fee percent
    qrVersion: int  # qr code version, 1 - nspk, 2 - tlv encoded, 3 - tlv plain
    rate: float  # exchange rate
    tlv: TLVExtended | None = None
    # merchant: Optional[str] = Field(default=None)  # merchant title
    # logo: Optional[str] = Field(default=None)  # merchant logo


class PayOutTlvRequest(BaseRequestModel):
    quoteId: str  # quote.id ref
    invoiceId: Optional[str] = Field(default=None)
    clientId: Optional[str] = Field(default=None)
    src_amount: Optional[float] = Field(default=None)
    sender_personal: Optional[PayOutSenderModel] = Field(default=None)



from __future__ import annotations
import logging

import json
from urllib.parse import urlencode
from pydantic import Field, BaseModel

from .types import TokenResponse
from .tokens import AccessToken, RefreshToken
from .exceptions import APIResponseError, APIAuthenticationError
from .models import (
    Credentials,
    GetQuoteModel,
    GetQuoteResponseModel,
    PayInModel,
    PayInResponseModel,
    PayOutModel,
    PayOutResponseModel,
    InvoiceModel,
    GetQuoteTlv,
    PayOutTlvRequest,
    QuoteTlvResponse,
)
from .enums import ApiPaths
from .transport import Request, Response
from .logger import Logger
from .cache import AbstractCache, DefaultCache

import httpx


class BaseClient:

    def __init__(self, config: Credentials, baseUrl: str, timeout: int = 20, debug: bool=False):
        self.config = config
        self.cache = DefaultCache()
        self.baseUrl = baseUrl
        self.timeout = timeout
        if debug:
            logging.basicConfig(level=logging.DEBUG)


class ApiAsyncClient(BaseClient):
    async def PayIn(self, request: PayInModel) -> PayInResponseModel:
         # Prepare request
        request = Request(
            method="post",
            path=ApiPaths.invoices_payin,
            content_type='application/json',
            noAuth=False,
            body=request.model_dump(exclude_none=True),
        )

        # Handle response
        response = await self._send_request(request)
        if (response.success):
            return response.cast(PayInResponseModel, APIResponseError)
        else:
            raise APIResponseError(response)

    async def PayOut(self, request: PayOutModel) -> PayOutResponseModel:
         # Prepare request
        request = Request(
            method="post",
            path=ApiPaths.invoices_payout,
            content_type='application/json',
            noAuth=False,
            signature=True,
            body=request.model_dump(exclude_none=True)
        )

        # Handle response
        response = await self._send_request(request)
        if (response.success):
            return response.cast(PayOutResponseModel, APIResponseError)
        else:
            raise APIResponseError(response)

    async def PayOutTlv(self, request: PayOutTlvRequest) -> PayOutResponseModel:
        request = Request(
            method="post",
            path=ApiPaths.invoices_payout_tlv,
            content_type="application/json",
            noAuth=False,
            signature=False,
            body=request.model_dump(exclude_none=True),
        )

        # Handle response
        response = await self._send_request(request)
        if not response.success:
            raise APIResponseError(response)

        return response.cast(PayOutResponseModel, APIResponseError)

    async def Quote(self, params: GetQuoteModel) -> GetQuoteResponseModel:
         # Prepare request
        request = Request(
            method="get",
            path=ApiPaths.fx_quote,
            content_type='application/json',
            noAuth=False,
            signature=False,
            body=params.model_dump(exclude_none=True)
        )

        # Handle response
        response = await self._send_request(request)
        if not response.success:
            raise APIResponseError(response)

        return response.cast(GetQuoteResponseModel, APIResponseError)

    async def QuoteQr(self, params: GetQuoteTlv) -> QuoteTlvResponse:
        request = Request(
            method="post",
            path=ApiPaths.fx_quote_tlv,
            content_type="application/json",
            noAuth=False,
            signature=False,
            body=params.model_dump(exclude_none=True),
        )

        # Handle response
        response = await self._send_request(request)
        if not response.success:
            raise APIResponseError(response)

        return response.cast(QuoteTlvResponse, APIResponseError)

    async def Status(self, id: str) -> InvoiceModel:
         # Prepare request
        request = Request(
            method="get",
            path=ApiPaths.invoices_info.replace(':id', id),
            content_type='application/json',
            noAuth=False,
            signature=False,
        )

        # Handle response
        response = await self._send_request(request)
        if not response.success:
            raise APIResponseError(response)
        return response.cast(InvoiceModel, APIResponseError)

    async def get_token(self) -> AccessToken | None:
        # First check if valid token is cached
        token = self.cache.get_token("AccessToken")
        refresh = self.cache.get_token("RefreshToken")

        if token is not None and not token.is_expired:
            return token
        else:
            # try to refresh token
            if refresh is not None and not refresh.is_expired:
                refreshed = await self._refresh_token(token, refresh)

                if refreshed.success:
                    access = AccessToken(refreshed.json_body["access_token"])
                    refresh = RefreshToken(
                        refreshed.json_body["refresh_token"],
                        int(refreshed.json_body["expires_in"]),
                    )

                    self.cache.set_token(access)
                    self.cache.set_token(refresh)

                    return access

            # try to issue token
            response = await self._fetch_token()
            if response.success:
                access = AccessToken(response.json_body["access_token"])
                refresh = RefreshToken(
                    response.json_body["refresh_token"],
                    int(response.json_body["expires_in"]),
                )

                self.cache.set_token(access)
                self.cache.set_token(refresh)

                return access
            else:
                raise APIAuthenticationError(response)

    async def _send_request(self, request: Request) -> Response:
        """
        Send a specified Request to the GoPay REST API and process the response
        """
        dict_factory = lambda l: {k: v for k, v in l if v is not None}
        body = request.body
        # Add Bearer authentication to headers if needed
        headers = request.headers or {}
        if not request.noAuth:
            auth = await self.get_token()
            if auth is not None:
                headers["Authorization"] = f"Bearer {auth.token}"

        client = httpx.AsyncClient(timeout=self.timeout)
        if request.method == 'get':
            url = f'{self.baseUrl}{request.path}'
            if body:
                params = urlencode(body)
                url = f'{url}?{params}'
            r = await client.request(
		        method=request.method,
                url=url,
                headers=headers,
                timeout=self.timeout
            )
        else:
            r = await client.request(
		        method=request.method,
 		        url=f"{self.baseUrl}{request.path}",
                headers=headers,
                json=body,
                timeout=self.timeout
            )

        # Build Response instance, try to decode body as JSON
        response = Response(raw_body=r.content, json={}, status_code=r.status_code)

        try:
            response.json_body = r.json()
        except json.JSONDecodeError:
            pass

        return response

    async def _fetch_token(self) -> Response:
        # Prepare request
        request = Request(
            method="post",
            path=ApiPaths.token_issue,
            content_type='application/json',
            noAuth=True,
            body={"account_id": self.config.account_id, "public_key": self.config.public_key},
        )
        # Handle response
        response = await self._send_request(request)
        return response
    
    async def _refresh_token(self) -> Response:
         # Prepare request
        request = Request(
            method="post",
            path=ApiPaths.token_refresh,
            content_type='application/json',
            body={"refresh_token": self.refreshToken},
        )
        # Handle response
        response = await self._send_request(request)
        return response


class ApiClient(BaseClient):

    def PayIn(self, request: PayInModel) -> PayInResponseModel:
         # Prepare request
        request = Request(
            method="post",
            path=ApiPaths.invoices_payin,
            content_type='application/json',
            noAuth=False,
            body=request.model_dump(exclude_none=True),
        )

        # Handle response
        response = self._send_request(request)
        if (response.success):
            return response.cast(PayInResponseModel, APIResponseError)
        else:
            raise APIResponseError(response)
        
    def PayOut(self, request: PayOutModel) -> PayOutResponseModel:
         # Prepare request
        request = Request(
            method="post",
            path=ApiPaths.invoices_payout,
            content_type='application/json',
            noAuth=False,
            signature=True,
            body=request.model_dump(exclude_none=True)
        )

        # Handle response
        response = self._send_request(request)
        if (response.success):
            return response.cast(PayOutResponseModel, APIResponseError)
        else:
            raise APIResponseError(response)

    def PayOutTlv(self, request: PayOutTlvRequest) -> PayOutResponseModel:
        request = Request(
            method="post",
            path=ApiPaths.invoices_payout_tlv,
            content_type="application/json",
            noAuth=False,
            signature=False,
            body=request.model_dump(exclude_none=True),
        )

        # Handle response
        response = self._send_request(request)
        if not response.success:
            raise APIResponseError(response)
        return response.cast(PayOutResponseModel, APIResponseError)

    def Quote(self, params: GetQuoteModel) -> GetQuoteResponseModel:
         # Prepare request
        request = Request(
            method="get",
            path=ApiPaths.fx_quote,
            content_type='application/json',
            noAuth=False,
            signature=False,
            body=params.model_dump(exclude_none=True)
        )

        # Handle response
        response = self._send_request(request)
        if not response.success:
            raise APIResponseError(response)

        return response.cast(GetQuoteResponseModel, APIResponseError)

    def QuoteQr(self, params: GetQuoteTlv) -> QuoteTlvResponse:
        request = Request(
            method="post",
            path=ApiPaths.fx_quote_tlv,
            content_type="application/json",
            noAuth=False,
            signature=False,
            body=params.model_dump(exclude_none=True),
        )

        # Handle response
        response = self._send_request(request)
        if not response.success:
            raise APIResponseError(response)

        return response.cast(QuoteTlvResponse, APIResponseError)

    def Status(self, id: str) -> InvoiceModel:
         # Prepare request
        request = Request(
            method="get",
            path=ApiPaths.invoices_info.replace(':id', id),
            content_type='application/json',
            noAuth=False,
            signature=False,
        )

        # Handle response
        response = self._send_request(request)
        if not response.success:
            raise APIResponseError(response)

        return response.cast(InvoiceModel, APIResponseError)

    def get_token(self) -> AccessToken | None:
        # First check if valid token is cached
        token = self.cache.get_token('access')
        refresh = self.cache.get_token('refresh')
        if token is not None and not token.is_expired:
            return token
        else:
            # try to refresh token
            if refresh is not None and not refresh.is_expired:
                refreshed = self._refresh_token()

                if (refreshed.success):
                    access = AccessToken(
                        response.json_body["access_token"]
                    )
                    refresh = RefreshToken(
                        response.json_body["refresh_token"],
                        int(response.json_body["expires_in"]),
                    )
                    self.cache.set_token(access)
                    self.cache.set_token(refresh)

                    return access

            # try to issue token
            response = self._fetch_token()
            if response.success:
                
                access = AccessToken(
                    response.json_body["access_token"]
                )
                refresh = RefreshToken(
                    response.json_body["refresh_token"],
                    int(response.json_body["expires_in"]),
                )
                self.cache.set_token(access)
                self.cache.set_token(refresh)

                return access
            else:
                raise APIAuthenticationError(response)

    def _send_request(self, request: Request) -> Response:
        """
        Send a specified Request to the GoPay REST API and process the response
        """
        # Add Bearer authentication to headers if needed
        headers = request.headers or {}
        if not request.noAuth:
            auth = self.get_token()
            if auth is not None:
                headers["Authorization"] = f"Bearer {auth.token}"

        if (request.method == 'get'):
            params = urlencode(body)
            r = httpx.request(
                method=request.method,
                url=f"{self.baseUrl}{request.path}?{params}",
                headers=headers,
                timeout=self.timeout
            )
        else:
            r = httpx.request(
                method=request.method,
                url=f"{self.baseUrl}{request.path}",
                headers=headers,
                json=body,
                timeout=self.timeout
            )

        # Build Response instance, try to decode body as JSON
        response = Response(raw_body=r.content, json={}, status_code=r.status_code)

        try:
            response.json_body = r.json()
        except json.JSONDecodeError:
            pass
        return response

    def _fetch_token(self) -> Response:
        # Prepare request
        request = Request(
            method="post",
            path=ApiPaths.token_issue,
            content_type='application/json',
            noAuth=True,
            body={"account_id": self.config.account_id, "public_key": self.config.public_key},
        )
        # Handle response
        response = self._send_request(request)
        return response

    def _refresh_token(self) -> Response:
         # Prepare request
        request = Request(
            method="post",
            path=ApiPaths.token_refresh,
            content_type='application/json',
            body={"refresh_token": self.refreshToken},
        )
        # Handle response
        response = self._send_request(request)
        return response



from abc import ABC, abstractmethod
from dataclasses import dataclass, field

from paymentsgate.tokens import AccessToken, RefreshToken


class AbstractCache(ABC):
    """
    Abstract class for implementing custom caches used to cache the token
    """

    @abstractmethod
    def get_token(self, key: str) -> AccessToken | RefreshToken:
        """
        Fetch a token with the specified key from the cache
        """
        ...

    @abstractmethod
    def set_token(self, token: AccessToken | RefreshToken) -> None:
        """
        Save the token to the cache under the specified key
        """
        ...


@dataclass
class DefaultCache(AbstractCache):
    tokens: dict[str, AccessToken | RefreshToken] = field(
        default_factory=dict, init=False
    )

    def get_token(self, key: str) -> AccessToken | RefreshToken | None:
        return self.tokens.get(key)

    def set_token(self, token: AccessToken | RefreshToken) -> None:
        self.tokens[token.__class__.__name__] = token



from paymentsgate.client import ApiClient
from paymentsgate.enums import (
    AuthenticationRealms,
    ApiPaths,
    Currencies,
    Languages,
    Statuses,
    CurrencyTypes,
    InvoiceTypes,
    CredentialsTypes,
    RiskScoreLevels,
    CancellationReason,
    FeesStrategy,
    InvoiceDirection,
    TTLUnits,
)
from paymentsgate.models import Credentials
